from __future__ import annotations
# from dataclasses import dataclass
from airtouch2.common.Buffer import Buffer

from airtouch2.common.interfaces import Serializable
from airtouch2.protocol.at2plus.constants import Limits
from airtouch2.protocol.at2plus.control_status_common import CONTROL_STATUS_SUBHEADER_LENGTH, ControlStatusSubHeader, ControlStatusSubType, SubDataLength
from airtouch2.protocol.at2plus.enums import GroupSetDamper, GroupSetPower
from airtouch2.protocol.at2plus.message_common import AddressMsgType, Header, MessageType, add_checksum_message_buffer, prime_message_buffer


GROUP_SETTINGS_LENGTH = 4


class GroupSettings(Serializable):
    id: int
    damp_mode: GroupSetDamper
    power: GroupSetPower
    damp: int | None

    def __init__(self, group_id: int, damp_mode: GroupSetDamper, power: GroupSetPower, damp: int | None = None):
        if damp is not None:
            if not 0 <= damp <= 100:
                raise ValueError(f"Damper percentage must be from 0 to 100")
        if not 0 <= group_id < Limits.MAX_GROUPS:
            raise ValueError(f'Group ID must be from 0 to {Limits.MAX_GROUPS-1}')
        self.id = group_id
        self.damp_mode = damp_mode
        self.power = power
        self.damp = damp

    def to_bytes(self) -> bytes:
        buffer = Buffer(GROUP_SETTINGS_LENGTH)
        buffer.append_bytes(
            bytes([
                self.id,
                (self.damp_mode << 5) | (self.power),
                self.damp if self.damp is not None else 255,
                0
            ])
        )
        return buffer.to_bytes()

    @staticmethod
    def from_bytes(repeat_data: bytes) -> GroupSettings:
        if (len(repeat_data) != GROUP_SETTINGS_LENGTH):
            raise ValueError(f"repeat_data must be {GROUP_SETTINGS_LENGTH} bytes")
        id = repeat_data[0] & 0x0F
        damp_mode = GroupSetDamper.from_int((repeat_data[1] >> 5) & 0x07)
        power = GroupSetPower.from_int((repeat_data[1]) & 0x07)
        damp = repeat_data[2] if 0 <= repeat_data[2] <= 100 else None

        return GroupSettings(id, damp_mode, power, damp)


class GroupControlMessage(Serializable):
    settings: list[GroupSettings]

    def __init__(self, settings: list[GroupSettings]):
        self.settings = settings

    def to_bytes(self) -> bytes:
        subheader = ControlStatusSubHeader(ControlStatusSubType.GROUP_CONTROL, SubDataLength(
            0, len(self.settings), GROUP_SETTINGS_LENGTH))
        buffer = prime_message_buffer(Header(AddressMsgType.NORMAL, MessageType.CONTROL_STATUS,
                                             CONTROL_STATUS_SUBHEADER_LENGTH + subheader.subdata_length.total()))
        buffer.append(subheader)
        for setting in self.settings:
            buffer.append(setting)
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()



from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum
from airtouch2.common.Buffer import Buffer

from airtouch2.common.interfaces import Serializable

SUBHEADER_MAGIC = 0xFF
EXTENDED_SUBHEADER_LENGTH = 2


class ExtendedMessageSubType(IntEnum):
    ERROR = 0x10
    ABILITY = 0x11
    GROUP_NAME = 0x12


@dataclass
class ExtendedSubHeader(Serializable):
    sub_type: ExtendedMessageSubType

    @staticmethod
    def from_bytes(subheader_bytes: bytes) -> ExtendedSubHeader:
        if len(subheader_bytes) != EXTENDED_SUBHEADER_LENGTH:
            raise ValueError("Unexpected control/status subheader size")
        if subheader_bytes[0] != SUBHEADER_MAGIC:
            raise ValueError("Subheader magic is incorrect")
        subtype = ExtendedMessageSubType(subheader_bytes[1])
        return ExtendedSubHeader(subtype)

    def to_bytes(self):
        return bytes([SUBHEADER_MAGIC, self.sub_type])

    @staticmethod
    def from_buffer(buffer: Buffer) -> ExtendedSubHeader:
        return ExtendedSubHeader.from_bytes(buffer.read_bytes(EXTENDED_SUBHEADER_LENGTH))



import asyncio
from datetime import datetime
import logging

from airtouch2.common.NetClient import NetClient
from airtouch2.protocol.at2.constants import MessageLength
from airtouch2.protocol.at2.messages import RequestState, SystemInfo
from airtouch2.at2.At2Aircon import At2Aircon
from airtouch2.at2.At2Group import At2Group
from airtouch2.common.interfaces import add_callback, Callback, Serializable, TaskCreator

_LOGGER = logging.getLogger(__name__)


class At2Client:
    aircons_by_id: dict[int, At2Aircon]
    groups_by_id: dict[int, At2Group]
    system_name: str

    def __init__(self, host: str, dump_responses: bool = False, task_creator: TaskCreator = asyncio.create_task):
        self.aircons_by_id = {}
        self.groups_by_id = {}
        self.system_name: str = "UNKNOWN"

        self._client = NetClient(host, 8899, self._on_connect, self._handle_one_message, task_creator)
        self._dump_responses: bool = dump_responses
        self._new_ac_callbacks: list[Callback] = []
        self._new_group_callbacks: list[Callback] = []
        self._found_ac = asyncio.Event()

        self.add_new_ac_callback(lambda: self._found_ac.set())

    async def connect(self) -> bool:
        return await self._client.connect()

    def run(self) -> None:
        self._client.run()

    async def wait_for_ac(self, timeout: int = 5) -> None:
        try:
            await asyncio.wait_for(self._found_ac.wait(), timeout)
        except TimeoutError:
            pass

    async def stop(self) -> None:
        await self._client.stop()

    def add_new_ac_callback(self, callback: Callback) -> Callback:
        """
        Subscribe 'callback' to new AC discoveries.
        Return a callback to unsubscribe.
        """
        return add_callback(callback, self._new_ac_callbacks)

    def add_new_group_callback(self, callback: Callback):
        """
        Subscribe 'callback' to new group discoveries.
        Return a callback to unsubscribe.
        """
        return add_callback(callback, self._new_group_callbacks)

    async def send(self, msg: Serializable):
        await self._client.send(msg)

    async def _on_connect(self):
        await self._client.send(RequestState())

    async def _read_response(self) -> SystemInfo | None:
        _LOGGER.debug("Waiting for response")
        resp = await self._client.read_bytes(MessageLength.RESPONSE)
        _LOGGER.debug("Got response")
        if not resp:
            return None

        if self._dump_responses:
            # blocks but is only used for dev and debugging
            with open('response_' + datetime.now().strftime("%m-%d-%Y_%H-%M-%S") + '.dump', 'wb') as f:
                f.write(resp)

        return SystemInfo.from_bytes(resp)

    async def _handle_one_message(self) -> None:
        system_info = await self._read_response()
        if not system_info:
            # something went wrong
            _LOGGER.info("Reading response message failed")
            return

        _LOGGER.debug(f"SystemInfo: {system_info}")
        # ACs
        for id, ac_info in system_info.aircons_by_id.items():
            if id not in self.aircons_by_id:
                self.aircons_by_id[id] = At2Aircon(self, ac_info)
                for callback in self._new_ac_callbacks:
                    callback()
            else:
                self.aircons_by_id[id].update(ac_info)

        # Groups
        for id, group_info in system_info.groups_by_id.items():
            if id not in self.groups_by_id:
                self.groups_by_id[id] = At2Group(self, group_info)
                for callback in self._new_group_callbacks:
                    callback()
            else:
                self.groups_by_id[id].update(group_info)



from abc import ABC, abstractmethod
from asyncio import Task
from typing import Awaitable, Callable, Coroutine, Protocol, TypeVar


class Serializable(ABC):
    @abstractmethod
    def to_bytes(self) -> bytes:
        pass


SendCoro = Callable[[Serializable], Awaitable[None]]
RecvCoro = Callable[[int], Awaitable[bytes | None]]
Callback = Callable[[], None]
CoroCallback = Callable[[], Awaitable[None]]
TaskCreator = Callable[[Coroutine], Task]


class Publisher(ABC):
    @abstractmethod
    def add_callback(self, callback: Callback) -> Callback:
        """Subscribe 'callback' to info updates. Return a callback that unsubscribes."""
        pass


# dumb thing required to pass containers of implementations as parameters
# to functions that expect containers of interfaces.
PublisherType = TypeVar("PublisherType", bound=Publisher)


def add_callback(callback: Callback, callbacks: list[Callback]) -> Callback:
    callbacks.append(callback)

    def remove_callback() -> None:
        if callback in callbacks:
            callbacks.remove(callback)

    return remove_callback



import asyncio
from datetime import datetime
import logging

from airtouch2.at2plus.At2PlusAircon import At2PlusAircon
from airtouch2.at2plus.At2PlusGroup import At2PlusGroup
from airtouch2.common.NetClient import NetClient
from airtouch2.protocol.at2plus.control_status_common import ControlStatusSubHeader, ControlStatusSubType
from airtouch2.protocol.at2plus.extended_common import ExtendedMessageSubType, ExtendedSubHeader
from airtouch2.protocol.at2plus.message_common import HEADER_LENGTH, HEADER_MAGIC, Header, Message, MessageType
from airtouch2.protocol.at2plus.messages.AcAbilityMessage import AcAbility, AcAbilityMessage, RequestAcAbilityMessage
from airtouch2.protocol.at2plus.messages.AcStatus import AcStatusMessage
from airtouch2.common.Buffer import Buffer
from airtouch2.protocol.at2plus.crc16_modbus import crc16
from airtouch2.common.interfaces import Callback, Serializable, TaskCreator
from airtouch2.protocol.at2plus.messages.GroupStatus import GroupStatusMessage

_LOGGER = logging.getLogger(__name__)


class At2PlusClient:
    def __init__(self, host: str, dump_responses: bool = False, task_creator: TaskCreator = asyncio.create_task):
        # public
        self.aircons_by_id: dict[int, At2PlusAircon] = {}
        self.groups_by_id: dict[int, At2PlusGroup] = {}

        # private
        self._client = NetClient(host, 9200, self._on_connect, self.handle_one_message, task_creator)
        self._dump_responses = dump_responses
        self._task_creator = task_creator
        self._new_ac_callbacks: list[Callback] = []
        self._ability_message_queue: asyncio.Queue[AcAbilityMessage] = asyncio.Queue()
        self._found_ac = asyncio.Event()
        self._new_group_callbacks: list[Callback] = []

        self.add_new_ac_callback(lambda: self._found_ac.set())

    async def connect(self) -> bool:
        return await self._client.connect()

    def run(self) -> None:
        self._client.run()

    async def wait_for_ac(self, timeout: int = 5) -> None:
        await asyncio.wait_for(self._found_ac.wait(), timeout)

    async def stop(self) -> None:
        await self._client.stop()

    def add_new_ac_callback(self, callback: Callback):
        self._new_ac_callbacks.append(callback)

        def remove_callback() -> None:
            if callback in self._new_ac_callbacks:
                self._new_ac_callbacks.remove(callback)

        return remove_callback

    def add_new_group_callback(self, callback: Callback):
        self._new_group_callbacks.append(callback)

        def remove_callback() -> None:
            if callback in self._new_group_callbacks:
                self._new_group_callbacks.remove(callback)

        return remove_callback

    async def send(self, msg: Serializable):
        await self._client.send(msg)

    async def handle_one_message(self) -> None:
        message = await self._read_message()
        if not message:
            # something went wrong
            _LOGGER.warning("Reading message failed")
            return

        if message.header.type == MessageType.CONTROL_STATUS:
            subheader = ControlStatusSubHeader.from_buffer(message.data_buffer)
            if subheader.sub_type == ControlStatusSubType.AC_STATUS:
                status_message = AcStatusMessage.from_bytes(
                    message.data_buffer.read_bytes(subheader.subdata_length.total()))
                self._task_creator(self._handle_status_message(status_message))
            elif subheader.sub_type == ControlStatusSubType.GROUP_STATUS:
                group_status_message = GroupStatusMessage.from_bytes(
                    message.data_buffer.read_bytes(subheader.subdata_length.total()))
                self._task_creator(self._handle_group_status_message(group_status_message))
            else:
                _LOGGER.warning(
                    f"Unknown status message type: subtype={subheader.sub_type}, data={message.data_buffer.to_bytes().hex(':')}")
        elif message.header.type == MessageType.EXTENDED:
            subheader = ExtendedSubHeader.from_buffer(message.data_buffer)
            if subheader.sub_type == ExtendedMessageSubType.ABILITY:
                ability_message_bytes = message.data_buffer.read_remaining()
                _LOGGER.debug(f"Creating ability message from {len(ability_message_bytes)} bytes")
                ability = AcAbilityMessage.from_bytes(ability_message_bytes)
                await self._ability_message_queue.put(ability)
            else:
                _LOGGER.warning(
                    f"Unknown extended message type: subtype={subheader.sub_type}, data={message.data_buffer.to_bytes().hex(':')}")
        else:
            _LOGGER.warning(
                f"Unknown message type, header={message.header.to_bytes().hex(':')}, data={message.data_buffer.to_bytes().hex(':')}")

    async def _read_magic(self) -> bytes:
        """Search for the two header magic bytes"""
        while True:  # exit via return on successful read of header magic
            byte = await self._client.read_bytes(1)
            while (byte is None or byte[0] != HEADER_MAGIC):
                byte = await self._client.read_bytes(1)

            byte = await self._client.read_bytes(1)
            if (byte is not None and byte[0] == HEADER_MAGIC):
                return bytes([HEADER_MAGIC, HEADER_MAGIC])

    async def _read_header(self) -> tuple[Header, bytes]:
        """Try to read header until successful. Returns the raw bytes as well for validating checksum"""
        while True:  # exit via return on successful read of header
            header_bytes = bytearray()
            header_bytes += await self._read_magic()
            header_remainder = await self._client.read_bytes(HEADER_LENGTH-2)

            if not header_remainder:
                _LOGGER.debug("Failed reading header, trying again")
                continue

            header_bytes += header_remainder
            try:
                header = Header.from_bytes(header_bytes)
                return (header, header_bytes)
            except ValueError as e:
                _LOGGER.debug(f"ValueError: {e}\nFailed reading header, trying again")

    async def _read_message(self) -> Message | None:
        "Try to read an entire message. Return None if reading was interrupted by network failure."
        header, header_bytes = await self._read_header()
        buffer = Buffer(header.data_length)

        data_bytes = await self._client.read_bytes(header.data_length)
        if not data_bytes:
            # interrupted during data reading
            return None
        if not buffer.append_bytes(data_bytes):
            _LOGGER.warning(
                f"Received incorrect number of bytes, expected {header.data_length} but received {buffer._head}")

        checksum = await self._client.read_bytes(2)
        if not checksum:
            # interrupted during checksum reading
            return None
        calculated_checksum = crc16(header_bytes[2:] + buffer._data)
        if (checksum != calculated_checksum):
            _LOGGER.warning(
                f"Checksum mismatch, ignoring message: Got {checksum.hex(':')}, expected {calculated_checksum.hex(':')}")
            return None

        if self._dump_responses:
            # blocks but is only used for dev and debugging
            with open('message_' + datetime.now().strftime("%m-%d-%Y_%H-%M-%S") + '.dump', 'wb') as f:
                f.write(header.to_bytes() + buffer.to_bytes() + checksum)

        return Message(header, buffer)

    async def _on_connect(self) -> None:
        # request groups
        await self._client.send(GroupStatusMessage([]))
        # request ACs
        await self._client.send(AcStatusMessage([]))

    async def _handle_status_message(self, message: AcStatusMessage):
        _LOGGER.debug("Handling AC status message")
        for status in message.statuses:
            if status.id not in self.aircons_by_id.keys():
                _LOGGER.debug(f"New AC ({status.id}) found")
                self.aircons_by_id[status.id] = At2PlusAircon(status, self)
                for callback in self._new_ac_callbacks:
                    callback()
                ability = await self._request_ac_ability(status.id)
                while not ability:
                    ability = await self._request_ac_ability(status.id)
                self.aircons_by_id[status.id]._set_ability(ability)
                _LOGGER.debug(f"Set ability of AC{status.id}")
            self.aircons_by_id[status.id]._update_status(status)
            _LOGGER.debug(f"Updated AC {status.id} with value {status}")
        _LOGGER.debug("Finished handling AC status message")

    async def _request_ac_ability(self, id: int) -> AcAbility | None:
        _LOGGER.debug(f"Requesting ability of AC{id}")
        await self._client.send(RequestAcAbilityMessage(id))
        _LOGGER.debug("Waiting for ability message response...")
        ac_ability = await self._ability_message_queue.get()
        _LOGGER.debug("Got ability message response")
        if len(ac_ability.abilities) != 1:
            _LOGGER.warning(f"Expected ability of single requested AC but got {len(ac_ability.abilities)}")
            return None
        if ac_ability.abilities[0].ac_id != id:
            _LOGGER.warning(f"Requested ability of AC{id} but got AC{ac_ability.abilities[0].ac_id}")
            return None
        _LOGGER.debug(f"Got ability of AC{id}: {ac_ability.abilities[0]}")
        return ac_ability.abilities[0]

    async def _handle_group_status_message(self, message: GroupStatusMessage):
        _LOGGER.debug("Handling group status message")
        for status in message.statuses:
            if status.id not in self.groups_by_id.keys():
                _LOGGER.debug(f"New group ({status.id}) found")
                self.groups_by_id[status.id] = At2PlusGroup(status, self)
                for callback in self._new_group_callbacks:
                    callback()
            self.groups_by_id[status.id]._update_status(status)
            _LOGGER.debug(f"Updated group {status.id} with value {status}")
        _LOGGER.debug("Finished handling group status message")




from __future__ import annotations
from enum import IntEnum
from airtouch2.protocol.at2plus.constants import Limits
from airtouch2.protocol.at2plus.conversions import setpoint_from_value, value_from_setpoint
from airtouch2.protocol.at2plus.control_status_common import CONTROL_STATUS_SUBHEADER_LENGTH, ControlStatusSubType, SubDataLength, ControlStatusSubHeader
from airtouch2.protocol.at2plus.enums import AcFanSpeed, AcSetMode, AcSetPower
from airtouch2.protocol.at2plus.message_common import AddressMsgType, Header, MessageType, add_checksum_message_buffer, prime_message_buffer
from airtouch2.common.interfaces import Serializable

AC_SETTINGS_LENGTH = 4


class SetpointControl(IntEnum):
    KEEP = 0
    CHANGE = 0x40


class AcSettings(Serializable):
    id: int
    power: AcSetPower
    mode: AcSetMode
    speed: AcFanSpeed
    setpoint: float | None

    def __init__(self, ac_id: int, power: AcSetPower, mode: AcSetMode, speed: AcFanSpeed, setpoint: float | None = None):
        if setpoint is not None:
            if not Limits.SETPOINT_MIN <= setpoint <= Limits.SETPOINT_MAX:
                raise ValueError(
                    f'Setpoint must be from {Limits.SETPOINT_MIN} to {Limits.SETPOINT_MAX}')
        if not 0 <= ac_id < Limits.MAX_ACS:
            raise ValueError(f'AC ID must be from 0 to {Limits.MAX_ACS-1}')
        self.id = ac_id
        self.power = power
        self.mode = mode
        self.speed = speed
        self.setpoint = setpoint

    def to_bytes(self) -> bytes:
        data = bytearray(AC_SETTINGS_LENGTH)
        data[0] = (self.power << 4) | self.id
        data[1] = (self.mode << 4) | self.speed
        data[2] = SetpointControl.CHANGE if self.setpoint is not None else SetpointControl.KEEP
        data[3] = value_from_setpoint(self.setpoint)
        return data

    @staticmethod
    def from_bytes(data: bytes) -> AcSettings:
        if (len(data) != AC_SETTINGS_LENGTH):
            raise ValueError(f"Data must be {AC_SETTINGS_LENGTH} bytes")
        id = data[0] & 0x0F
        power = AcSetPower.from_int((data[0] & 0xF0) >> 4)
        mode = AcSetMode.from_int((data[1] & 0xF0) >> 4)
        speed = AcFanSpeed.from_int(data[1] & 0x0F)
        change_setpoint = data[2] == SetpointControl.CHANGE
        setpoint = setpoint_from_value(data[3]) if change_setpoint else None
        return AcSettings(id, power, mode, speed, setpoint)


class AcControlMessage(Serializable):
    """AcControl Message to control all ACs"""
    settings: list[AcSettings]

    def __init__(self, settings: list[AcSettings]):
        self.settings = settings

    def to_bytes(self) -> bytes:
        subheader = ControlStatusSubHeader(ControlStatusSubType.AC_CONTROL, SubDataLength(
            0, len(self.settings), AC_SETTINGS_LENGTH))
        buffer = prime_message_buffer(Header(AddressMsgType.NORMAL, MessageType.CONTROL_STATUS,
                                      CONTROL_STATUS_SUBHEADER_LENGTH + subheader.subdata_length.total()))
        buffer.append(subheader)
        for setting in self.settings:
            buffer.append(setting)
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()



from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum
from airtouch2.common.Buffer import Buffer
import logging

from airtouch2.common.interfaces import Serializable


CONTROL_STATUS_SUBHEADER_LENGTH = 8
SUBDATALENGTH_LENGTH = 6

_LOGGER = logging.getLogger(__name__)

class ControlStatusOffsets(IntEnum):
    # Control/Status messages 'data' consists of:
    # * 8 byte header with information as described below, immediately followed by:
    # * a 'normal data', which occurs once, with length described in the header.
    # * a number of 'repeat data', which each have length described by the header
    # 1st byte is one of ControlStatusSubType
    SUBTYPE = 0
    # 2nd byte empty
    # bytes 3-4 are 'normal data' length
    NORMAL_DATA_LENGTH = 2
    # bytes 5-6 are each 'repeat data' length
    REPEAT_DATA_LENGTH = 4
    # bytes 7-8 are number of repeated 'repeat data's
    REPEAT_DATA_COUNT = 6
    SUBDATA = 8


class ControlStatusSubType(IntEnum):
    UNSET = 0
    GROUP_CONTROL = 0x20
    GROUP_STATUS = 0x21
    AC_CONTROL = 0x22
    AC_STATUS = 0x23


@dataclass
class SubDataLength(Serializable):
    normal: int
    repeat_count: int
    repeat_length: int

    def total(self):
        return self.normal + self.repeat_count * self.repeat_length

    @staticmethod
    def from_bytes(length_bytes: bytes) -> SubDataLength:
        if len(length_bytes) != SUBDATALENGTH_LENGTH:
            raise ValueError("Unexpected SubDataLength size")
        normal = int.from_bytes(length_bytes[0:2], 'big')
        repeat_length = int.from_bytes(length_bytes[2:4], 'big')
        repeat_count = int.from_bytes(length_bytes[4:6], 'big')
        return SubDataLength(normal, repeat_count, repeat_length)

    def to_bytes(self) -> bytes:
        return self.normal.to_bytes(2, 'big') + \
            self.repeat_length.to_bytes(2, 'big') + \
            self.repeat_count.to_bytes(2, 'big')


@dataclass
class ControlStatusSubHeader(Serializable):
    sub_type: ControlStatusSubType
    subdata_length: SubDataLength

    @staticmethod
    def from_bytes(subheader_bytes: bytes) -> ControlStatusSubHeader:
        if len(subheader_bytes) != CONTROL_STATUS_SUBHEADER_LENGTH:
            raise ValueError("Unexpected control/status subheader size")
        try:
            subtype = ControlStatusSubType(subheader_bytes[0])
        except ValueError as e:
            _LOGGER.warning(
                f"Unknown message type in header ({hex(subheader_bytes[0])})", exc_info=e)
            subtype = ControlStatusSubType.UNSET

        data_length = SubDataLength.from_bytes(subheader_bytes[2:8])
        return ControlStatusSubHeader(subtype, data_length)

    @staticmethod
    def from_buffer(buffer: Buffer) -> ControlStatusSubHeader:
        return ControlStatusSubHeader.from_bytes(buffer.read_bytes(CONTROL_STATUS_SUBHEADER_LENGTH))

    def to_bytes(self) -> bytes:
        return self.sub_type.to_bytes(1, 'big') + \
            bytes(1) + \
            self.subdata_length.to_bytes()



from airtouch2.protocol.at2.messages.SystemInfo import SystemInfo
from airtouch2.protocol.at2.messages.RequestState import RequestState
from airtouch2.protocol.at2.messages.ac_commands import ChangeSetTemperature, ToggleAc, SetFanSpeed, SetMode
from airtouch2.protocol.at2.messages.group_commands import ToggleGroup, ChangeDamper



import asyncio
import errno
import logging
from socket import gaierror
from typing import Callable
from airtouch2.common.interfaces import CoroCallback, Serializable, TaskCreator

_LOGGER = logging.getLogger(__name__)

NetworkOrHostDownErrors = (errno.EHOSTUNREACH, errno.ECONNREFUSED,  errno.ETIMEDOUT,
                           errno.ENETDOWN, errno.ENETUNREACH, errno.ENETRESET, errno.ECONNABORTED)


class NetClient:
    """A generic network client"""

    def __init__(self, host: str, port: int, on_connect: CoroCallback, handle_message: CoroCallback,
                 task_creator: TaskCreator = asyncio.create_task):
        # network
        self._host_ip: str = host
        self._host_port: int = port
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None

        # async
        self._task_creator: Callable = task_creator
        self._main_loop_task: asyncio.Task[None] | None = None
        self._stop: bool = False

        self._on_connect = on_connect
        self._handle_message = handle_message

    async def connect(self) -> bool:
        """Opens connection to the server, returns True/False if successful/unsuccessful"""
        _LOGGER.debug(f"Connecting to {self._host_ip} on port {self._host_port}")
        try:
            self._reader, self._writer = await asyncio.open_connection(self._host_ip, self._host_port)
        except OSError as e:
            _LOGGER.warning(f"Could not connect to host {self._host_ip}")
            if isinstance(e, gaierror):
                # provided ip or port is rubbish/invalid
                pass
            elif e.errno not in NetworkOrHostDownErrors:
                raise e
            return False
        else:
            await self._on_connect()
            return True

    def run(self) -> None:
        """Starts the processing of incoming information from the server"""
        _LOGGER.debug("Starting listener task")
        self._main_loop_task = self._task_creator(self._main())

    async def stop(self) -> None:
        """Stops the processing of incoming information from the server"""
        if not self._main_loop_task:
            raise RuntimeError("Client task is not running")
        self._stop = True
        self._main_loop_task.cancel()
        try:
            await self._main_loop_task
        except asyncio.CancelledError as e:
            # Eat the expected exception
            pass

    async def send(self, message: Serializable) -> None:
        """Send the serializable 'message'"""
        if self._writer is None:
            raise RuntimeError("Client is not connected - call connect() first")
        else:
            bytes_to_write = message.to_bytes()
            _LOGGER.debug(f"Sending {message.__class__.__name__} with data: {bytes_to_write.hex(':')}")
            _LOGGER.debug(f"{repr(message)}")
            self._writer.write(bytes_to_write)
            drained: bool = False
            while not drained:
                try:
                    await self._writer.drain()
                    drained = True
                except (ConnectionResetError, asyncio.IncompleteReadError) as e:
                    await self._try_reconnect()

    async def read_bytes(self, size: int) -> bytes | None:
        """
        Read exactly 'size' bytes, return None if could not read enough bytes or on disconnection and reconnection.
        This coroutine handles reconnection.
        """
        if self._reader is None:
            raise RuntimeError("Client is not connected - call connect() first")
        try:
            data = await self._reader.readexactly(size)
        except asyncio.IncompleteReadError as e:
            _LOGGER.debug(f"IncompleteReadError - partial bytes: {e.partial.hex(':')}")
            data = None
        except ConnectionResetError as e:
            _LOGGER.debug(f"ConnectionResetError")
            data = None

        if data is None:
            _LOGGER.warning("Connection lost, reconnecting")
            await self._try_reconnect()
            return None
        _LOGGER.debug(f"Read payload of size {size}: {data.hex(':')}")
        return data

    async def _main(self) -> None:
        while not self._stop:
            if not (self._reader and self._writer):
                raise RuntimeError("Client is not connected - call connect() first")
            await self._handle_message()

    async def _try_reconnect(self) -> None:
        retries = 0
        while not await self.connect():
            await asyncio.sleep(0.001 * (10**retries) if retries < 4 else 10)
            retries += 1
            if not retries % 60 or retries == 4:
                _LOGGER.info("Server is not responding, will continue trying to reconnect every 10s")
        _LOGGER.info("Reconnected")